from typing import List, Dict, Tuple, Optional, Union, Set, TypeVar
from common import State, Character, Eps, Terminal, Nonterminal, Emptyset
from dfa import DFA
from nfa import NFA
from grammars import RegGrammar
from regex import RegEx, AST, Bin, Iter, BinOp, IterOp, CharNode
import antlr4 # type: ignore
from DFALexer import DFALexer
from DFAParser import DFAParser
from DFAListener import DFAListener
from NFALexer import NFALexer
from NFAParser import NFAParser
from NFAListener import NFAListener
from RegExLexer import RegExLexer
from RegExParser import RegExParser
from RegExVisitor import RegExVisitor

class ParsingError(Exception):
    def __init__(self, args):
        self.args = args

class Parser:

    def __init__(self):
        pass

    def names_to_str(self, collection: Union[Set[State], Set[Character], Set[Terminal], Set[Nonterminal]]) -> str:
        return "{" + ','.join(set(map(lambda x: x.name, collection))) + "}"

    def reggrammar_to_str(self, reg: RegGrammar, full: bool = False) -> str:
        rules = self.rules_to_str(reg.rules)
        if not full:
            return "P=" + rules

        # full - verbose description of DFA - only for development, dismiss later
        nonterminals = self.names_to_str(reg.nonterminals)
        terminals = self.names_to_str(reg.terminals)

        out = "REG = (" + nonterminals + ", " + terminals + ", P," + reg.init.name + ")\nP="
        out += (self.rules_to_str(reg.rules))
        return out

    def rules_to_str(self, rules: RegGrammar.Rules) -> str:
        out = ""
        for rule in rules:
            out = out[:-1] + (",\n")
            rewrite = rules[rule]
            if len(rewrite) == 0:
                continue

            out += (rule.name + "->")
            for variant in rewrite:
                if isinstance(variant, Terminal):
                    out += variant.name
                else:
                    for i in range(len(variant)):
                        out += variant[i].name
                if next:  # TODO
                    out += "|"

        return "{" + out[2:len(out) - 1] + "}"

    def dfa_to_str(self, dfa: DFA, full : bool = False) -> str:
        transition = ""
        for key, state_2 in dfa.transition.items():
            state_1, character = key
            transition += "(" + state_1.name + "," + character.name + ")=" + state_2.name + " "

        init = "init=" + dfa.init.name
        final = "final=" + self.names_to_str(dfa.final)

        # full - verbose description of DFA - only for development, dismiss later
        if full:
            return "DFA = (" + self.names_to_str(dfa.states) + ", " + self.names_to_str(dfa.characters) + \
                   ", d, " + init + ", " + final + ")\n" + transition
        else:
            return init + " " + transition + final

    def nfa_to_str(self, nfa: NFA, full : bool = False) -> str:
        transition = ""
        for key, set_states in nfa.transition.items():
            state, character = key
            char = character.name if not isinstance(character, Eps) else "ε"
            dest_states = nfa.transition[state, character]
            transition += "(" + state.name + "," + char + ")=" + self.names_to_str(dest_states) + " "

        init = "init=" + nfa.init.name
        final = "final=" + self.names_to_str(nfa.final)

        if full:
            return "NFA = (" + self.names_to_str(nfa.states) + ", " + self.names_to_str(nfa.characters) + \
                   ", d, " + init + ", " + final + ")\n" + transition
        else:
            return init + " " + transition + " " + final

    def regex_to_str(self, reg: RegEx) -> str:
        return reg.expression.astprint()

    def common_parse(self, string: str, given_lexer, given_parser, given_listener):
        chars = antlr4.InputStream(string)
        lexer = given_lexer(chars)
        tokens = antlr4.CommonTokenStream(lexer)
        parser = given_parser(tokens)

        tree = parser.start()
        listener = given_listener()
        walker = antlr4.ParseTreeWalker()
        walker.walk(listener, tree)
        return listener

    def str_to_dfa(self, string: str) -> DFA:
        listener = self.common_parse(string, DFALexer, DFAParser, DFAListener)

        if listener.init is None:
            listener.init = listener.first_state
            if listener.init is None:
                raise ParsingError("Automat musí obsahovat alespoň jeden stav.")

        dfa = DFA(listener.states, listener.characters, listener.transition, listener.init, listener.final)
        return dfa

    def str_to_nfa(self, string: str) -> NFA:
        listener = self.common_parse(string, NFALexer, NFAParser, NFAListener)

        if listener.init is None:
            listener.init = listener.first_state
            if listener.init is None:
                raise ParsingError("Automat musí obsahovat alespoň jeden stav.")

        return NFA(listener.states, listener.characters, listener.transition, listener.init, listener.final)

    def str_to_regex(self, string: str) -> RegEx:
        try:
            chars = antlr4.InputStream(string)
            lexer = RegExLexer(chars)
            tokens = antlr4.CommonTokenStream(lexer)
            parser = RegExParser(tokens)

            tree = parser.start()
            ast = RegExVisitor()
            ast.visitStart(tree)

            return RegEx(ast.characters, ast.expression)

        except Exception as e:
            raise ParsingError(e.args)


class DFAListener(DFAListener):
    def __init__(self):
        self.states = set()
        self.characters = set()
        self.transition = {}
        self.init = None
        self.first_state = None
        self.final = set()

    def exitInit(self, ctx):
        if ctx.statename() is not None:
            state = State(str(ctx.statename().STATE()))
            self.init = state
            self.states.add(state)

    def exitProduction(self, ctx):
        state = State(str(ctx.statename(0).STATE()))
        character = Character(str(ctx.statename(1).STATE()))
        dest_state = State(str(ctx.statename(2).STATE()))
        self.states.add(state)
        self.states.add(dest_state)
        self.characters.add(character)
        self.transition[state, character] = dest_state

        if self.first_state is None:
            self.first_state = state

    def exitFinal(self, ctx):
        i = 0
        while ctx.statename(i) is not None:
            state = State(str(ctx.statename(i).STATE()))
            self.states.add(state)
            self.final.add(state)
            i += 1

class NFAListener(NFAListener):
    def __init__(self):
        self.states = set()
        self.characters = set()
        self.transition = {}
        self.init = None
        self.first_state = None
        self.final = set()
        self.efa = False

    def exitInit(self, ctx):
        state = State(str(ctx.statename().STATE()))
        self.init = state
        self.states.add(state)

    def exitProduction(self, ctx):
        state = State(str(ctx.statename(0).STATE()))
        self.states.add(state)

        dest_states = set()
        i = 0
        while ctx.stateset().statename(i) is not None:
            dest_state = State(str(ctx.stateset().statename(i).STATE()))
            self.states.add(dest_state)
            dest_states.add(dest_state)
            i += 1

        if ctx.EPSILON():
            self.transition[state, Eps()] = dest_states
            self.efa = True
        else:
            character = Character(str(ctx.statename(1).STATE()))
            self.characters.add(character)
            self.transition[state, character] = dest_states

        if self.first_state is None:
            self.first_state = state

    def exitFinal(self, ctx):
        i = 0
        while ctx.stateset().statename(i) is not None:
            state = State(str(ctx.stateset().statename(i).STATE()))
            self.states.add(state)
            self.final.add(state)
            i += 1


class RegExVisitor(RegExVisitor):
    def __init__(self):
        self.characters: Set[Character] = set()
        self.expression: AST = None

    def visitStart(self, ctx):
        self.expression = self.visitExpr(ctx.expr())

    def visitExpr(self, ctx):
        # Binary operation: union or explicit concatenation
        if ctx.UNION() or ctx.CONCAT():
            op = Bin.Union if ctx.UNION() is not None else Bin.Concat
            return BinOp(self.visitExpr(ctx.expr(0)), op, self.visitExpr(ctx.expr(1)))

        # Implicit concatenation of (iterated) symbols or expressions in parentheses
        expressions = list(map(lambda x: self.visitConcatenable(x), ctx.concatenated()))
        return self.implicit_concat(expressions)

    def visitConcatenable(self, ctx):
        if ctx.symbol():
            return self.visitSymbol(ctx.symbol())

        elif ctx.iterable():
            if ctx.ITER():
                return self.visitIterable(ctx.iterable())

            elif ctx.POS_ITER():
                return self.visitIterable(ctx.iterable(), True)

        elif ctx.parentheses():
            return self.visitParentheses(ctx.parentheses())

    def visitSymbol(self, ctx):
        if ctx.ALPHABET():
            self.characters.add(Character(str(ctx.ALPHABET())))
            return CharNode(Character(str(ctx.ALPHABET())))

        elif ctx.EPSILON():
            return CharNode(Eps())

        elif ctx.EMPTYSET():
            return CharNode(Emptyset())

    def visitParentheses(self, ctx):
        return self.visitExpr(ctx.expr())

    def visitIterable(self, ctx, positive=False):
        if ctx.symbol():
            expression = self.visitSymbol(ctx.symbol())

        elif ctx.parentheses():
            expression = self.visitParentheses(ctx.parentheses())

        return IterOp(expression, Iter.Positive) if positive else IterOp(expression, Iter.Iteration)

    def implicit_concat(self, to_concat):
        ast = to_concat[0]
        if len(to_concat) > 1:
            for expression in to_concat[1:]:
                ast = BinOp(ast, Bin.Concat, expression)
        return ast

